from canalyst_candas.candas.func_music import FuncMusic
from canalyst_candas.candas.chart import Chart
from canalyst_candas.candas.modelset import ModelSet
from canalyst_candas.candas.model import Model, ModelMap
from canalyst_candas.candas.search import Search
from canalyst_candas.candas.sample_model import SampleModel


def help():
    """
    Canalyst Candas help function
    """
    print("Canalyst Candas help")
    print(
        "Please go to https://pypi.org/project/canalyst-candas/#description for installation help"
    )
    print("For support, please contact candas@canalyst.com")
    print("For an API KEY please go to https://app.canalyst.com/u/settings/api-tokens")
    print("For an Excel model download:get_excel_model(ticker, config)")
